// WL_DRAW.C

#include "wl_def.h"
#pragma hdrstop

#include "wl_cloudsky.h"
#include "wl_atmos.h"
#include "wl_shade.h"

/*
=============================================================================

                               LOCAL CONSTANTS

=============================================================================
*/

// the door is the last picture before the sprites
#define DOORWALL        (PMSpriteStart-8)

#define ACTORSIZE       0x4000

/*
=============================================================================

                              GLOBAL VARIABLES

=============================================================================
*/

static byte *vbuf = NULL;
unsigned vbufPitch = 0;

int32_t    lasttimecount;
int32_t    frameon;
boolean fpscounter;

int fps_frames=0, fps_time=0, fps=0;

int *wallheight;
int min_wallheight;

//
// math tables
//
short *pixelangle;
int32_t finetangent[FINEANGLES/4];
fixed sintable[ANGLES+ANGLES/4];
fixed *costable = sintable+(ANGLES/4);

//
// refresh variables
//
fixed   viewx,viewy;                    // the focal point
short   viewangle;
fixed   viewsin,viewcos;

void    TransformActor (objtype *ob);
void    BuildTables (void);
void    ClearScreen (void);
int     CalcRotate (objtype *ob);
void    DrawScaleds (void);
void    CalcTics (void);
void    ThreeDRefresh (void);



//
// wall optimization variables
//
int     lastside;               // true for vertical
int32_t    lastintercept;
int     lasttilehit;
int     lasttexture;

//
// ray tracing variables
//
short    focaltx,focalty,viewtx,viewty;
longword xpartialup,xpartialdown,ypartialup,ypartialdown;

short   midangle,angle;

word    tilehit;
int     pixx;

short   xtile,ytile;
short   xtilestep,ytilestep;
int32_t    xintercept,yintercept;
word    xspot,yspot;
int     texdelta;
boolean setvis;
boolean inupperwall;

#define MAXSPANS 32

typedef enum
{
  SPAN_NORMAL,
  SPAN_UPPER,
  SPAN_UPPER_END
} span_type;

typedef struct _span_t
{
  span_type type;
  int height;
  int texturePage;
  int textureOffset;
} span_t;

int numSpans;
span_t spans[MAXSPANS];

word horizwall[MAXWALLTILES],vertwall[MAXWALLTILES];

/*
============================================================================

                           3 - D  DEFINITIONS

============================================================================
*/

/*
========================
=
= TransformActor
=
= Takes paramaters:
=   gx,gy               : globalx/globaly of point
=
= globals:
=   viewx,viewy         : point of view
=   viewcos,viewsin     : sin/cos of viewangle
=   scale               : conversion from global value to screen value
=
= sets:
=   screenx,transx,transy,screenheight: projected edge location and size
=
========================
*/
void TransformActor (objtype *ob)
{
  fixed gx,gy,gxt,gyt,nx,ny;
// translate point to view centered coordinates
  gx = ob->x-viewx;
  gy = ob->y-viewy;
// calculate newx
  gxt = FixedMul(gx,viewcos);
  gyt = FixedMul(gy,viewsin);
  nx = gxt-gyt-ACTORSIZE;         // fudge the shape forward a bit, because
                                  // the midpoint could put parts of the shape
                                  // into an adjacent wall
// calculate newy
  gxt = FixedMul(gx,viewsin);
  gyt = FixedMul(gy,viewcos);
  ny = gyt+gxt;
// calculate perspective ratio
  ob->transx = nx;
  ob->transy = ny;
  if (nx<MINDIST)                 // too close, don't overflow the divide
  {
    ob->viewheight = 0;
    return;
  }
  ob->viewx = (word)(centerx + ny*scale/nx);
// calculate height (heightnumerator/(nx>>8))
  ob->viewheight = (word)(heightnumerator/(nx>>8));
}

//==========================================================================

/*
========================
=
= TransformTile
=
= Takes paramaters:
=   tx,ty               : tile the object is centered in
=
= globals:
=   viewx,viewy         : point of view
=   viewcos,viewsin     : sin/cos of viewangle
=   scale               : conversion from global value to screen value
=
= sets:
=   screenx,transx,transy,screenheight: projected edge location and size
=
= Returns true if the tile is withing getting distance
=
========================
*/
boolean TransformTile (int tx, int ty, short *dispx, short *dispheight)
{
  fixed gx,gy,gxt,gyt,nx,ny;
// translate point to view centered coordinates
  gx = ((int32_t)tx<<TILESHIFT)+0x8000-viewx;
  gy = ((int32_t)ty<<TILESHIFT)+0x8000-viewy;
// calculate newx
  gxt = FixedMul(gx,viewcos);
  gyt = FixedMul(gy,viewsin);
  nx = gxt-gyt-0x2000;            // 0x2000 is size of object
// calculate newy
  gxt = FixedMul(gx,viewsin);
  gyt = FixedMul(gy,viewcos);
  ny = gyt+gxt;
// calculate height / perspective ratio
  if (nx<MINDIST) *dispheight = 0; // too close, don't overflow the divide
  else
  {
    *dispx = (short)(centerx + ny*scale/nx);
    *dispheight = (short)(heightnumerator/(nx>>8));
  }
// see if it should be grabbed
  if (nx<TILEGLOBAL && ny>-TILEGLOBAL/2 && ny<TILEGLOBAL/2) return true;
  return false;
}

//==========================================================================

/*
====================
=
= CalcHeight
=
= Calculates the height of xintercept,yintercept from viewx,viewy
=
====================
*/
int CalcHeight()
{
  fixed z = FixedMul(xintercept - viewx, viewcos)
          - FixedMul(yintercept - viewy, viewsin);
  if(z < MINDIST) z = MINDIST;
  int height = heightnumerator / (z >> 8);
  if(height < min_wallheight) min_wallheight = height;
  return height;
}

int CalcHeight(int32_t x, int32_t y)
{
  fixed z = FixedMul(x - viewx, viewcos) - FixedMul(y - viewy, viewsin);
  if(z < MINDIST) z = MINDIST;
  int height = heightnumerator / (z >> 8);
  if(height < min_wallheight) min_wallheight = height;
  return height;
}

//==========================================================================

/*
===================
=
= ScalePost
=
===================
*/
byte *postsource;
int postx;
int postwidth;
void ScalePost()
{
  int ywcount, yoffs, yw, yd, yendoffs;
  byte col;
  return;

#ifdef USE_SHADING
  byte *curshades = shadetable[GetShade(wallheight[postx])];
#endif

  ywcount = yd = wallheight[postx] >> 3;
  if(yd <= 0) yd = 100;
  yoffs = (viewheight / 2 - ywcount) * vbufPitch;
  if(yoffs < 0) yoffs = 0;
  yoffs += postx;

  yendoffs = viewheight / 2 + ywcount - 1;
  yw=TEXTURESIZE-1;

  while(yendoffs >= viewheight)
  {
    ywcount -= TEXTURESIZE/2;
    while(ywcount <= 0)
    {
      ywcount += yd;
      yw--;
    }
    yendoffs--;
  }
  if(yw < 0) return;
#ifdef USE_SHADING
  col = curshades[postsource[yw]];
#else
  col = postsource[yw];
#endif
  yendoffs = yendoffs * vbufPitch + postx;
  while(yoffs <= yendoffs)
  {
    vbuf[yendoffs] = col;
    ywcount -= TEXTURESIZE/2;
    if(ywcount <= 0)
    {
      do
      {
        ywcount += yd;
        yw--;
      }
      while(ywcount <= 0);
      if(yw < 0) break;
#ifdef USE_SHADING
      col = curshades[postsource[yw]];
#else
      col = postsource[yw];
#endif
    }
    yendoffs -= vbufPitch;
  }
}

void GlobalScalePost(byte *vidbuf, unsigned pitch)
{
  vbuf = vidbuf;
  vbufPitch = pitch;
  ScalePost();
}

static void DrawSpan(span_t *span)
{
  int ywcount, yoffs, yw, yd, yendoffs;
  byte col;
  byte *texture;

#ifdef USE_SHADING
  byte *curshades = shadetable[GetShade(span->height)];
#endif

  ywcount = yd = span->height >> 3;
  if(yd <= 0) yd = 100;

  if (span->type == SPAN_NORMAL)
  {
    yoffs = (viewheight / 2 - ywcount) * vbufPitch;
    yendoffs = viewheight / 2 + ywcount - 1;
  }
  else
  {
    yoffs = (viewheight / 2 - ((3 * span->height) >> 3)) * vbufPitch;
    yendoffs = viewheight / 2 - ywcount - 1;
  }
  if(yoffs < 0) yoffs = 0;
  yoffs += pixx;
  yw=TEXTURESIZE-1;
  while(yendoffs >= viewheight)
  {
    ywcount -= TEXTURESIZE/2;
    while(ywcount <= 0)
    {
      ywcount += yd;
      yw--;
    }
    yendoffs--;
  }
  if(yw < 0) return;
  texture = PM_GetPage(span->texturePage) + span->textureOffset;
#ifdef USE_SHADING
  col = curshades[texture[yw]];
#else
  col = texture[yw];
#endif
  yendoffs = yendoffs * vbufPitch + pixx;
  while(yoffs <= yendoffs)
  {
    vbuf[yendoffs] = col;
    ywcount -= TEXTURESIZE/2;
    if(ywcount <= 0)
    {
      do
      {
        ywcount += yd;
        yw--;
      }
      while(ywcount <= 0);
      if(yw < 0) break;
#ifdef USE_SHADING
      col = curshades[texture[yw]];
#else
      col = texture[yw];
#endif
    }
    yendoffs -= vbufPitch;
  }
}

static void DrawCeiling(int height1, int height2)
{
  int y = viewheight / 2 - (height1 >> 3);
  if (y < 0) y = 0;
  y = y * vbufPitch + pixx;
  int yend = (viewheight / 2 - (height2 >> 3)) * vbufPitch + pixx;
  while(y < yend)
  {
    vbuf[y] = 4;
    y += vbufPitch;
  }
}

static void DrawSpans()
{
  int i;
  int lastUpperEnd = 0;
  boolean wallheightset = false;
  for(i = numSpans - 1; i >= 0; i--)
  {
    if (!wallheightset && spans[i].type == SPAN_NORMAL)
    {
      wallheight[pixx] = spans[i].height;
      wallheightset = true;
    }
    if(spans[i].type == SPAN_UPPER_END)
    {
      lastUpperEnd = spans[i].height;
      continue;
    }
    else if(spans[i].type == SPAN_UPPER && lastUpperEnd)
    {
      DrawCeiling(spans[i].height, lastUpperEnd);
      lastUpperEnd = 0;
    }
    DrawSpan(&spans[i]);
  }
  if (!wallheightset) wallheight[pixx] = 0;
}

/*
====================
=
= HitVertWall
=
= tilehit bit 7 is 0, because it's not a door tile
= if bit 6 is 1 and the adjacent tile is a door tile, use door side pic
=
====================
*/
void HitVertWall (void)
{
  int wallpic;
  int texture;

  texture = ((yintercept+texdelta)>>TEXTUREFROMFIXEDSHIFT)&TEXTUREMASK;
  if (xtilestep == -1)
  {
    texture = TEXTUREMASK-texture;
    xintercept += TILEGLOBAL;
  }

  if(lastside==1 && lastintercept==xtile && lasttilehit==tilehit && !(lasttilehit & 0x40))
  {
    if((pixx&3) && texture == lasttexture)
    {
      ScalePost();
      postx = pixx;
      wallheight[pixx] = wallheight[pixx-1];
      return;
    }
    ScalePost();
    wallheight[pixx] = CalcHeight();
    postsource+=texture-lasttexture;
    postwidth=1;
    postx=pixx;
    lasttexture=texture;
    return;
  }
  if(lastside!=-1) ScalePost();

  lastside=1;
  lastintercept=xtile;
  lasttilehit=tilehit;
  lasttexture=texture;
  wallheight[pixx] = CalcHeight();
  postx = pixx;
  postwidth = 1;

  if (tilehit & 0x40) // check for adjacent doors
  {
    ytile = (short)(yintercept>>TILESHIFT);
    if ( tilemap[xtile-xtilestep][ytile]&0x80 )
      wallpic = DOORWALL+3;
    else
      wallpic = vertwall[tilehit & ~0x40];
  }
  else wallpic = vertwall[tilehit];
  postsource = PM_GetTexture(wallpic) + texture;
}


/*
====================
=
= HitHorizWall
=
= tilehit bit 7 is 0, because it's not a door tile
= if bit 6 is 1 and the adjacent tile is a door tile, use door side pic
=
====================
*/

void HitHorizWall (void)
{
  int wallpic;
  int texture;

  texture = ((xintercept+texdelta)>>TEXTUREFROMFIXEDSHIFT)&TEXTUREMASK;
  if (ytilestep == -1)
    yintercept += TILEGLOBAL;
  else
    texture = TEXTUREMASK-texture;

  if(lastside==0 && lastintercept==ytile && lasttilehit==tilehit && !(lasttilehit & 0x40))
  {
    if((pixx&3) && texture == lasttexture)
    {
      ScalePost();
      postx=pixx;
      wallheight[pixx] = wallheight[pixx-1];
      return;
    }
    ScalePost();
    wallheight[pixx] = CalcHeight();
    postsource+=texture-lasttexture;
    postwidth=1;
    postx=pixx;
    lasttexture=texture;
    return;
  }
  if(lastside!=-1) ScalePost();

  lastside=0;
  lastintercept=ytile;
  lasttilehit=tilehit;
  lasttexture=texture;
  wallheight[pixx] = CalcHeight();
  postx = pixx;
  postwidth = 1;

  if (tilehit & 0x40)
  {                                                               // check for adjacent doors
    xtile = (short)(xintercept>>TILESHIFT);
    if ( tilemap[xtile][ytile-ytilestep]&0x80)
      wallpic = DOORWALL+2;
    else
      wallpic = horizwall[tilehit & ~0x40];
  }
  else wallpic = horizwall[tilehit];
  postsource = PM_GetTexture(wallpic) + texture;
}

//==========================================================================

/*
====================
=
= HitHorizDoor
=
====================
*/

void HitHorizDoor (void)
{
  int doorpage;
  int doornum;
  int texture;

  doornum = tilehit&0x7f;
  texture = ((xintercept-doorposition[doornum])>>TEXTUREFROMFIXEDSHIFT)&TEXTUREMASK;

  if(lasttilehit==tilehit)
  {
    if((pixx&3) && texture == lasttexture)
    {
      ScalePost();
      postx=pixx;
      wallheight[pixx] = wallheight[pixx-1];
      return;
    }
    ScalePost();
    wallheight[pixx] = CalcHeight();
    postsource+=texture-lasttexture;
    postwidth=1;
    postx=pixx;
    lasttexture=texture;
    return;
  }

  if(lastside!=-1) ScalePost();

  lastside=2;
  lasttilehit=tilehit;
  lasttexture=texture;
  wallheight[pixx] = CalcHeight();
  postx = pixx;
  postwidth = 1;

  switch(doorobjlist[doornum].lock)
  {
    case dr_normal:   doorpage = DOORWALL;   break;
    case dr_lock1:
    case dr_lock2:
    case dr_lock3:
    case dr_lock4:    doorpage = DOORWALL+6; break;
    case dr_elevator: doorpage = DOORWALL+4; break;
  }
  postsource = PM_GetTexture(doorpage) + texture;
}

//==========================================================================

/*
====================
=
= HitVertDoor
=
====================
*/
void HitVertDoor (void)
{
  int doorpage;
  int doornum;
  int texture;

  doornum = tilehit&0x7f;
  texture = ((yintercept-doorposition[doornum])>>TEXTUREFROMFIXEDSHIFT)&TEXTUREMASK;

  if(lasttilehit==tilehit)
  {
    if((pixx&3) && texture == lasttexture)
    {
      ScalePost();
      postx=pixx;
      wallheight[pixx] = wallheight[pixx-1];
      return;
    }
    ScalePost();
    wallheight[pixx] = CalcHeight();
    postsource+=texture-lasttexture;
    postwidth=1;
    postx=pixx;
    lasttexture=texture;
    return;
  }
  if(lastside!=-1) ScalePost();

  lastside=2;
  lasttilehit=tilehit;
  lasttexture=texture;
  wallheight[pixx] = CalcHeight();
  postx = pixx;
  postwidth = 1;

  switch(doorobjlist[doornum].lock)
  {
    case dr_normal:   doorpage = DOORWALL+1; break;
    case dr_lock1:
    case dr_lock2:
    case dr_lock3:
    case dr_lock4:    doorpage = DOORWALL+7; break;
    case dr_elevator: doorpage = DOORWALL+5; break;
  }
  postsource = PM_GetTexture(doorpage) + texture;
}

//==========================================================================

#define HitHorizBorder HitHorizWall
#define HitVertBorder HitVertWall

//==========================================================================

byte vgaCeiling[]=
{
#ifndef SPEAR
 0x1d,0x1d,0x1d,0x1d,0x1d,0x1d,0x1d,0x1d,0x1d,0xbf,
 0x4e,0x4e,0x4e,0x1d,0x8d,0x4e,0x1d,0x2d,0x1d,0x8d,
 0x1d,0x1d,0x1d,0x1d,0x1d,0x2d,0xdd,0x1d,0x1d,0x98,

 0x1d,0x9d,0x2d,0xdd,0xdd,0x9d,0x2d,0x4d,0x1d,0xdd,
 0x7d,0x1d,0x2d,0x2d,0xdd,0xd7,0x1d,0x1d,0x1d,0x2d,
 0x1d,0x1d,0x1d,0x1d,0xdd,0xdd,0x7d,0xdd,0xdd,0xdd
#else
 0x6f,0x4f,0x1d,0xde,0xdf,0x2e,0x7f,0x9e,0xae,0x7f,
 0x1d,0xde,0xdf,0xde,0xdf,0xde,0xe1,0xdc,0x2e,0x1d,0xdc
#endif
};

/*
=====================
=
= VGAClearScreen
=
=====================
*/
void VGAClearScreen (void)
{
  byte ceiling=vgaCeiling[gamestate.episode*10+mapon];
  int y;
  byte *ptr = vbuf;

#ifdef USE_SHADING
    for(y = 0; y < viewheight / 2; y++, ptr += vbufPitch)
        memset(ptr, shadetable[GetShade((viewheight / 2 - y) << 3)][ceiling], viewwidth);
    for(; y < viewheight; y++, ptr += vbufPitch)
        memset(ptr, shadetable[GetShade((y - viewheight / 2) << 3)][0x19], viewwidth);
#else
    for(y = 0; y < viewheight / 2; y++, ptr += vbufPitch)
        memset(ptr, ceiling, viewwidth);
    for(; y < viewheight; y++, ptr += vbufPitch)
        memset(ptr, 0x19, viewwidth);
#endif
}

//==========================================================================

/*
=====================
=
= CalcRotate
=
=====================
*/
int CalcRotate (objtype *ob)
{
  int angle, viewangle;

  // this isn't exactly correct, as it should vary by a trig value,
  // but it is close enough with only eight rotations
  viewangle = player->angle + (centerx - ob->viewx)/8;
  if (ob->obclass == rocketobj || ob->obclass == hrocketobj)
    angle = (viewangle-180) - ob->angle;
  else
    angle = (viewangle-180) - dirangle[ob->dir];

  angle+=ANGLES/16;
  while (angle>=ANGLES) angle-=ANGLES;
  while (angle<0) angle+=ANGLES;
  if (ob->state->rotate == 2)             // 2 rotation pain frame
    return 0;                             // pain with shooting frame bugfix
  return angle/(ANGLES/8);
}

void ScaleShape (int xcenter, int shapenum, unsigned height, uint32_t flags)
{
  t_compshape *shape;
  unsigned scale,pixheight;
  unsigned starty,endy;
  word *cmdptr;
  byte *cline;
  byte *line;
  byte *vmem;
  int actx,i,upperedge;
  short newstart;
  int scrstarty,screndy,lpix,rpix,pixcnt,ycnt;
  unsigned j;
  byte col;

#ifdef USE_SHADING
  byte *curshades;
  if(flags & FL_FULLBRIGHT)
    curshades = shadetable[0];
  else
    curshades = shadetable[GetShade(height)];
#endif
  shape = (t_compshape *) PM_GetSprite(shapenum);
  scale=height>>3;                 // low three bits are fractional
  if(!scale) return;   // too close or far away
  pixheight=scale*SPRITESCALEFACTOR;
  actx=xcenter-scale;
  upperedge=viewheight/2-scale;
  cmdptr=(word *) shape->dataofs;

  for(i=shape->leftpix,pixcnt=i*pixheight,rpix=(pixcnt>>6)+actx;i<=shape->rightpix;i++,cmdptr++)
  {
    lpix=rpix;
    if(lpix>=viewwidth) break;
    pixcnt+=pixheight;
    rpix=(pixcnt>>6)+actx;
    if(lpix!=rpix && rpix>0)
    {
      if(lpix<0) lpix=0;
      if(rpix>viewwidth) rpix=viewwidth,i=shape->rightpix+1;
      cline=(byte *)shape + *cmdptr;
      while(lpix<rpix)
      {
        if(wallheight[lpix]<=(int)height)
        {
          line=cline;
          while((endy = READWORD(line)) != 0)
          {
            endy >>= 1;
            newstart = READWORD(line);
            starty = READWORD(line) >> 1;
            j=starty;
            ycnt=j*pixheight;
            screndy=(ycnt>>6)+upperedge;
            if(screndy<0)
              vmem=vbuf+lpix;
            else
              vmem=vbuf+screndy*vbufPitch+lpix;
            for(;j<endy;j++)
            {
              scrstarty=screndy;
              ycnt+=pixheight;
              screndy=(ycnt>>6)+upperedge;
              if(scrstarty!=screndy && screndy>0)
              {
#ifdef USE_SHADING
                col=curshades[((byte *)shape)[newstart+j]];
#else
                col=((byte *)shape)[newstart+j];
#endif
                if(scrstarty<0) scrstarty=0;
                if(screndy>viewheight) screndy=viewheight,j=endy;
                while(scrstarty<screndy)
                {
                  *vmem=col;
                  vmem+=vbufPitch;
                  scrstarty++;
                }
              }
            }
          }
        }
        lpix++;
      }
    }
  }
}

void SimpleScaleShape (int xcenter, int shapenum, unsigned height)
{
  t_compshape   *shape;
  unsigned scale,pixheight;
  unsigned starty,endy;
  word *cmdptr;
  byte *cline;
  byte *line;
  int actx,i,upperedge;
  short newstart;
  int scrstarty,screndy,lpix,rpix,pixcnt,ycnt;
  unsigned j;
  byte col;
  byte *vmem;

  shape = (t_compshape *) PM_GetSprite(shapenum);

  scale=height>>1;
  pixheight=scale*SPRITESCALEFACTOR;
  actx=xcenter-scale;
  upperedge=viewheight/2-scale;
  cmdptr=shape->dataofs;
  for(i=shape->leftpix,pixcnt=i*pixheight,rpix=(pixcnt>>6)+actx;i<=shape->rightpix;i++,cmdptr++)
  {
    lpix=rpix;
    if(lpix>=viewwidth) break;
    pixcnt+=pixheight;
    rpix=(pixcnt>>6)+actx;
    if(lpix!=rpix && rpix>0)
    {
      if(lpix<0) lpix=0;
      if(rpix>viewwidth) rpix=viewwidth,i=shape->rightpix+1;
      cline = (byte *)shape + *cmdptr;
      while(lpix<rpix)
      {
        line=cline;
        while((endy = READWORD(line)) != 0)
        {
          endy >>= 1;
          newstart = READWORD(line);
          starty = READWORD(line) >> 1;
          j=starty;
          ycnt=j*pixheight;
          screndy=(ycnt>>6)+upperedge;
          if(screndy<0) vmem=vbuf+lpix;
          else vmem=vbuf+screndy*vbufPitch+lpix;
          for(;j<endy;j++)
          {
            scrstarty=screndy;
            ycnt+=pixheight;
            screndy=(ycnt>>6)+upperedge;
            if(scrstarty!=screndy && screndy>0)
            {
              col=((byte *)shape)[newstart+j];
              if(scrstarty<0) scrstarty=0;
              if(screndy>viewheight) screndy=viewheight,j=endy;
              while(scrstarty<screndy)
              {
                *vmem=col;
                vmem+=vbufPitch;
                scrstarty++;
              }
            }
          }
        }
        lpix++;
      }
    }
  }
}

/*
=====================
=
= DrawScaleds
=
= Draws all objects that are visable
=
=====================
*/
#define MAXVISABLE 250

typedef struct
{
  short      viewx,
             viewheight,
             shapenum;
  short      flags;          // this must be changed to uint32_t, when you
                               // you need more than 16-flags for drawing
#ifdef USE_DIR3DSPR
  statobj_t *transsprite;
#endif
} visobj_t;

visobj_t vislist[MAXVISABLE];
visobj_t *visptr,*visstep,*farthest;

void DrawScaleds (void)
{
  int      i,least,numvisable,height;
  byte     *tilespot,*visspot;
  unsigned spotloc;

  statobj_t *statptr;
  objtype   *obj;

  visptr = &vislist[0];
// place static objects
  for (statptr = &statobjlist[0] ; statptr !=laststatobj ; statptr++)
  {
    if ((visptr->shapenum = statptr->shapenum) == -1) continue; // object has been deleted
    if (!*statptr->visspot) continue;  // not visable
    if (TransformTile (statptr->tilex,statptr->tiley,&visptr->viewx,&visptr->viewheight) && statptr->flags & FL_BONUS)
    {
      GetBonus (statptr);
      if(statptr->shapenum == -1) continue; // object has been taken
    }
    if (!visptr->viewheight) continue; // to close to the object
#ifdef USE_DIR3DSPR
    if(statptr->flags & FL_DIR_MASK)
      visptr->transsprite=statptr;
    else
      visptr->transsprite=NULL;
#endif
    if (visptr < &vislist[MAXVISABLE-1])    // don't let it overflow
    {
      visptr->flags = (short) statptr->flags;
      visptr++;
    }
  }
  // place active objects
  for (obj = player->next;obj;obj=obj->next)
  {
    if ((visptr->shapenum = obj->state->shapenum)==0) continue; // no shape
    spotloc = (obj->tilex<<mapshift)+obj->tiley;   // optimize: keep in struct?
    visspot = &spotvis[0][0]+spotloc;
    tilespot = &tilemap[0][0]+spotloc;
    // could be in any of the nine surrounding tiles
    if (*visspot ||
       (*(visspot-1)  && !*(tilespot-1))  || (*(visspot+1)  && !*(tilespot+1)) ||
       (*(visspot-65) && !*(tilespot-65)) || (*(visspot-64) && !*(tilespot-64)) ||
       (*(visspot-63) && !*(tilespot-63)) || (*(visspot+65) && !*(tilespot+65)) ||
       (*(visspot+64) && !*(tilespot+64)) || (*(visspot+63) && !*(tilespot+63)))
    {
      obj->active = ac_yes;
      TransformActor (obj);
      if (!obj->viewheight) continue; // too close or far away
      visptr->viewx = obj->viewx;
      visptr->viewheight = obj->viewheight;
      if (visptr->shapenum == -1) visptr->shapenum = obj->temp1;  // special shape
      if (obj->state->rotate)     visptr->shapenum += CalcRotate (obj);
      if (visptr < &vislist[MAXVISABLE-1])    // don't let it overflow
      {
        visptr->flags = (short) obj->flags;
#ifdef USE_DIR3DSPR
        visptr->transsprite = NULL;
#endif
        visptr++;
      }
      obj->flags |= FL_VISABLE;
    }
    else
      obj->flags &= ~FL_VISABLE;
  }
// draw from back to front
  numvisable = (int) (visptr-&vislist[0]);
  if (!numvisable) return;   // no visable objects
  for (i = 0; i<numvisable; i++)
  {
    least = 32000;
    for (visstep=&vislist[0] ; visstep<visptr ; visstep++)
    {
      height = visstep->viewheight;
      if (height < least)
      {
        least = height;
        farthest = visstep;
      }
    }
    // draw farthest
#ifdef USE_DIR3DSPR
    if(farthest->transsprite) Scale3DShape(vbuf, vbufPitch, farthest->transsprite);
    else
#endif
    ScaleShape(farthest->viewx, farthest->shapenum, farthest->viewheight, farthest->flags);
    farthest->viewheight = 32000;
  }
}

//==========================================================================

/*
==============
=
= DrawPlayerWeapon
=
= Draw the player's hands
=
==============
*/
int weaponscale[NUMWEAPONS] = {SPR_KNIFEREADY, SPR_PISTOLREADY, SPR_MACHINEGUNREADY, SPR_CHAINREADY};

void DrawPlayerWeapon (void)
{
  int shapenum;
#ifndef SPEAR
  if (gamestate.victoryflag)
  {
#ifndef APOGEE_1_0
    if (player->state == &s_deathcam && (GetTimeCount()&32) )
      SimpleScaleShape(viewwidth/2,SPR_DEATHCAM,viewheight+1);
#endif
    return;
  }
#endif
  if (gamestate.weapon != -1)
  {
    shapenum = weaponscale[gamestate.weapon]+gamestate.weaponframe;
    SimpleScaleShape(viewwidth/2,shapenum,viewheight+1);
  }
  if (demorecord || demoplayback) SimpleScaleShape(viewwidth/2,SPR_DEMO,viewheight+1);
}


//==========================================================================


/*
=====================
=
= CalcTics
=
=====================
*/
void CalcTics (void)
{
  // calculate tics since last refresh for adaptive timing
  if (lasttimecount > (int32_t) GetTimeCount()) lasttimecount = GetTimeCount();    // if the game was paused a LONG time

  uint32_t curtime = SDL_GetTicks();
  tics = (curtime * 7) / 100 - lasttimecount;
  if(!tics)
  {
    // wait until end of current tic
    SDL_Delay(((lasttimecount + 1) * 100) / 7 - curtime);
    tics = 1;
  }
  lasttimecount += tics;
  if (tics>MAXTICS) tics = MAXTICS;
}

//==========================================================================

static void AddSolidSpan(int height, span_type type, int page, int offs)
{
  span_t *cur;

  if (numSpans >= MAXSPANS) return;
  if(page < 0 || page >= ChunksInFile) printf(".");
  cur = &spans[numSpans++];
  cur->type = type;
  cur->height = height;
  cur->texturePage = page;
  cur->textureOffset = offs;
}

static void AddSolidSpan_Front(span_type type, int page, boolean vertical, boolean flip)
{
  int32_t xglobal, yglobal;
  int texture;
  if (vertical)
  {
    xglobal = xtile << TILESHIFT;
    yglobal = yintercept;
    texture = ((yglobal + texdelta) >> TEXTUREFROMFIXEDSHIFT) & TEXTUREMASK;
    if (flip)
    {
      xglobal += TILEGLOBAL;
      texture = TEXTUREMASK - texture;
    }
  }
  else
  {
    xglobal = xintercept;
    yglobal = ytile << TILESHIFT;
    texture = ((xglobal + texdelta) >> TEXTUREFROMFIXEDSHIFT) & TEXTUREMASK;
    if (flip) yglobal += TILEGLOBAL;
    else texture = TEXTUREMASK - texture;
  }
  AddSolidSpan(CalcHeight(xglobal, yglobal), type, page, texture);
}

static boolean CheckIntersection(short xtile, short ytile, int32_t xglobal, int32_t yglobal, int dx, int dy, boolean vertical, boolean flip, boolean other)
{
  int wallpic;

  tilehit = tilemap[xtile][ytile];
  int ntile = tilehit & TM_N_MASK;
  int htile = tilehit & TM_H_MASK;
  if (!inupperwall && htile == TM_H_SOLID)
  {
    wallpic = MAPSPOT(xtile, ytile, 0) >> 8;
    if (wallpic) wallpic = (wallpic - 1) * 2 + vertical;
    else wallpic = vertical;
    AddSolidSpan_Front(SPAN_UPPER, wallpic, vertical, flip);
    inupperwall = true;
    if (!setvis) return true;
  }
  else if (inupperwall && (htile == TM_H_NONE || htile == TM_H_NONE_BLOCK))
  {
    AddSolidSpan_Front(SPAN_UPPER_END, 0, vertical, flip);
    inupperwall = false;
  }
  if (!setvis) return htile != TM_N_NONE;
  if (ntile == TM_N_NONE) return false;
  if (ntile == TM_N_PUSHWALL)
  {
    int texture;
    if (vertical)
    {
      int32_t yintbuf = yglobal + ((dy * pwallpos) >> 6);
      if ((yintbuf >> 16) != (yglobal >> 16)) return false;
      xglobal = xtile << TILESHIFT;
      yglobal = yintbuf;
      texture = (yglobal >> TEXTUREFROMFIXEDSHIFT) & TEXTUREMASK;
      if (flip)
      {
        xglobal += TILEGLOBAL - (pwallpos << 10);
        texture = TEXTUREMASK - texture;
      }
      else xglobal += pwallpos << 10;
    }
    else
    {
      int32_t xintbuf = xglobal + ((dx * pwallpos) >> 6);
      if ((xintbuf >> 16) != (xglobal >> 16)) return false;
      xglobal = xintbuf;
      yglobal = ytile << TILESHIFT;
      texture = (xglobal >> TEXTUREFROMFIXEDSHIFT) & TEXTUREMASK;
      if (flip) yglobal += TILEGLOBAL - (pwallpos << 10);
      else
      {
        texture = TEXTUREMASK - texture;
        yglobal += pwallpos << 10;
      }
    }
    wallpic = (pwalltile - 1) * 2 + vertical;
    int height = CalcHeight(xglobal, yglobal);
    AddSolidSpan(height, SPAN_NORMAL, wallpic, texture);
    if (htile == TM_H_NONE)
    {
      setvis = false;
      return false;   // transparent upper wall
    }
    if (inupperwall) AddSolidSpan(height, SPAN_UPPER_END, 0, 0);
    return true;
  }
  else if(ntile == TM_N_DOOR)
  {
    int doornum = MAPSPOT(xtile, ytile, 0) & 0x3f;
    int texture;

    if (vertical)
    {
      int32_t yintbuf=yglobal+(dy>>1);
      if((yintbuf>>16)!=(yglobal>>16)) return false;
      if((word)yintbuf<doorposition[doornum]) return false;
      yglobal=yintbuf;
      xglobal=(xtile<<TILESHIFT)|0x8000;
      ytile = (short) (yglobal >> TILESHIFT);
      texture = ((yglobal - doorposition[doornum]) >> TEXTUREFROMFIXEDSHIFT) & TEXTUREMASK;
    }
    else
    {
      int32_t xintbuf=xglobal+(dx>>1);
      if((xintbuf>>16)!=(xglobal>>16)) return false;
      if((word)xintbuf<doorposition[doornum]) return false;
      xglobal=xintbuf;
      yglobal=(ytile<<TILESHIFT)+0x8000;
      xtile = (short) (xglobal >> TILESHIFT);
      texture = ((xglobal - doorposition[doornum]) >> TEXTUREFROMFIXEDSHIFT) & TEXTUREMASK;
    }
    switch(doorobjlist[doornum].lock)
    {
      case dr_normal:
      default:          wallpic = DOORWALL; break;
      case dr_lock1:
      case dr_lock2:
      case dr_lock3:
      case dr_lock4:    wallpic = DOORWALL+6; break;
      case dr_elevator: wallpic = DOORWALL+4; break;
    }
    if (vertical) wallpic++;
    int height = CalcHeight(xglobal, yglobal);
    AddSolidSpan(height, SPAN_NORMAL, wallpic, texture);

    if (htile == TM_H_NONE)
    {
      setvis = false;
      return false;   // transparent upper wall
    }
    if (inupperwall) AddSolidSpan(height, SPAN_UPPER_END, 0, 0);
    return true;
  }
  else if (ntile == TM_N_SOLID)
  {
    // check for adjacent doors
    int adjxtile = xtile + (flip ? 1 : -1);
    int adjytile = ytile + (flip ? 1 : -1);
    int adjtile;

    if (vertical) adjtile = tilemap[adjxtile][ytile];
    else adjtile = tilemap[xtile][adjytile];

    if ((adjtile & TM_N_MASK) == TM_N_DOOR) wallpic = DOORWALL + 2 + vertical;
    else wallpic = ((MAPSPOT(xtile, ytile, 0) & 0xff) - 1) * 2 + vertical;
    AddSolidSpan_Front(SPAN_NORMAL, wallpic, vertical, flip);
  }
  if (htile == TM_H_NONE) // transparent upper wall
  {
    setvis = false;
    return false;
  }
  if (inupperwall) AddSolidSpan_Front(SPAN_UPPER_END, 0, vertical, flip);
  return true;
}

void AsmRefresh()
{
  int32_t xstep,ystep;
  longword xpartial,ypartial;
  byte playertile = tilemap[focaltx][focalty];
  boolean playerInPushwallBackTile = (playertile & TM_N_MASK) == TM_N_PUSHWALL;
  boolean playerUnderUpperWall = (playertile & TM_H_MASK) == TM_N_SOLID;

  for(pixx=0;pixx<viewwidth;pixx++)
  {
    short angl=midangle+pixelangle[pixx];
    if(angl<0) angl+=FINEANGLES;
    if(angl>=3600) angl-=FINEANGLES;
    if(angl<900)
    {
      xtilestep=1;
      ytilestep=-1;
      xstep=finetangent[900-1-angl];
      ystep=-finetangent[angl];
      xpartial=xpartialup;
      ypartial=ypartialdown;
    }
    else if(angl<1800)
    {
      xtilestep=-1;
      ytilestep=-1;
      xstep=-finetangent[angl-900];
      ystep=-finetangent[1800-1-angl];
      xpartial=xpartialdown;
      ypartial=ypartialdown;
    }
    else if(angl<2700)
    {
      xtilestep=-1;
      ytilestep=1;
      xstep=-finetangent[2700-1-angl];
      ystep=finetangent[angl-1800];
      xpartial=xpartialdown;
      ypartial=ypartialup;
    }
    else if(angl<3600)
    {
      xtilestep=1;
      ytilestep=1;
      xstep=finetangent[angl-2700];
      ystep=finetangent[3600-1-angl];
      xpartial=xpartialup;
      ypartial=ypartialup;
    }

    yintercept=FixedMul(ystep,xpartial)+viewy;
    xtile=focaltx+xtilestep;
    xspot=(word)((xtile<<mapshift)+((uint32_t)yintercept>>16));
    xintercept=FixedMul(xstep,ypartial)+viewx;
    ytile=focalty+ytilestep;
    yspot=(word)((((uint32_t)xintercept>>16)<<mapshift)+ytile);
    texdelta=0;

    // Special treatment when player is in back tile of pushwall
    if (playerInPushwallBackTile)
    {
      if (pwalldir == di_east && xtilestep ==  1 || pwalldir == di_west && xtilestep == -1)
      {
        int32_t yintbuf = yintercept - ((ystep * (64 - pwallpos)) >> 6);
        if((yintbuf >> 16) == focalty)   // ray hits pushwall back?
        {
          if(pwalldir == di_east)
            xintercept = (focaltx << TILESHIFT) + (pwallpos << 10);
          else
            xintercept = (focaltx << TILESHIFT) - TILEGLOBAL + ((64 - pwallpos) << 10);
          yintercept = yintbuf;
          ytile = (short) (yintercept >> TILESHIFT);
          tilehit = pwalltile;
          HitVertWall();
          continue;
        }
      }
      else if (pwalldir == di_south && ytilestep ==  1 ||  pwalldir == di_north && ytilestep == -1)
      {
        int32_t xintbuf = xintercept - ((xstep * (64 - pwallpos)) >> 6);
        if((xintbuf >> 16) == focaltx)   // ray hits pushwall back?
        {
          xintercept = xintbuf;
          if(pwalldir == di_south)
            yintercept = (focalty << TILESHIFT) + (pwallpos << 10);
          else
            yintercept = (focalty << TILESHIFT) - TILEGLOBAL + ((64 - pwallpos) << 10);
          xtile = (short) (xintercept >> TILESHIFT);
          tilehit = pwalltile;
          HitHorizWall();
          continue;
        }
      }
    }

    setvis = true;
    inupperwall = playerUnderUpperWall;
    numSpans = 0;

    if (inupperwall) AddSolidSpan(viewheight << 2, SPAN_UPPER, 1, 0);  // dummy, won't be painted

    do
    {
      if (ytilestep== -1 && (yintercept>>16)<=ytile) goto horizentry;
      if (ytilestep==  1 && (yintercept>>16)>=ytile) goto horizentry;
vertentry:
      if((uint32_t)yintercept>mapheight*65536-1 || (word)xtile>=mapwidth)
      {
        if(xtile<0) xintercept=0, xtile=0;
        else if(xtile>=mapwidth) xintercept=mapwidth<<TILESHIFT, xtile=mapwidth-1;
        else xtile=(short) (xintercept >> TILESHIFT);
        if(yintercept<0) yintercept=0, ytile=0;
        else if(yintercept>=(mapheight<<TILESHIFT)) yintercept=mapheight<<TILESHIFT, ytile=mapheight-1;
        yspot=0xffff;
        tilehit=0;
        HitHorizBorder();
        break;
      }
      if(xspot>=maparea) break;

      if(CheckIntersection(xtile, yintercept >> TILESHIFT, xintercept, yintercept, xstep, ystep, true, xtilestep == -1, ytilestep == -1)) break;

      if(setvis) *((byte *)spotvis+xspot)=1;
      xtile+=xtilestep;
      yintercept+=ystep;
      xspot=(word)((xtile<<mapshift)+((uint32_t)yintercept>>16));
    }
    while(1);

    DrawSpans();
    continue;

    do
    {
      if(xtilestep== -1 && (xintercept>>16)<=xtile) goto vertentry;
      if(xtilestep==  1 && (xintercept>>16)>=xtile) goto vertentry;
horizentry:
      if((uint32_t)xintercept>mapwidth*65536-1 || (word)ytile>=mapheight)
      {
        if(ytile<0) yintercept=0, ytile=0;
        else if(ytile>=mapheight) yintercept=mapheight<<TILESHIFT, ytile=mapheight-1;
        else ytile=(short) (yintercept >> TILESHIFT);
        if(xintercept<0) xintercept=0, xtile=0;
        else if(xintercept>=(mapwidth<<TILESHIFT)) xintercept=mapwidth<<TILESHIFT, xtile=mapwidth-1;
        xspot=0xffff;
        tilehit=0;
        HitVertBorder();
        break;
      }
      if(yspot>=maparea) break;

      if(CheckIntersection(xintercept >> TILESHIFT, ytile, xintercept, yintercept, xstep, ystep, false, ytilestep == -1, xtilestep == -1)) break;

      if (setvis) *((byte *)spotvis+yspot)=1;
      ytile+=ytilestep;
      xintercept+=xstep;
      yspot=(word)((((uint32_t)xintercept>>16)<<mapshift)+ytile);
    }
    while(1);
    DrawSpans();
  }
}

/*
====================
=
= WallRefresh
=
====================
*/
void WallRefresh (void)
{
  xpartialdown = viewx&(TILEGLOBAL-1);
  xpartialup = TILEGLOBAL-xpartialdown;
  ypartialdown = viewy&(TILEGLOBAL-1);
  ypartialup = TILEGLOBAL-ypartialdown;
  min_wallheight = viewheight;
  lastside = -1;                  // the first pixel is on a new wall
  AsmRefresh ();
  ScalePost ();                   // no more optimization on last post
}

void CalcViewVariables()
{
  viewangle = player->angle;
  midangle = viewangle*(FINEANGLES/ANGLES);
  viewsin = sintable[viewangle];
  viewcos = costable[viewangle];
  viewx = player->x - FixedMul(focallength,viewcos);
  viewy = player->y + FixedMul(focallength,viewsin);
  focaltx = (short)(viewx>>TILESHIFT);
  focalty = (short)(viewy>>TILESHIFT);
  viewtx = (short)(player->x >> TILESHIFT);
  viewty = (short)(player->y >> TILESHIFT);
}

//==========================================================================

/*
========================
=
= ThreeDRefresh
=
========================
*/

void    ThreeDRefresh (void)
{
//
// clear out the traced array
//
    memset(spotvis,0,maparea);
    spotvis[player->tilex][player->tiley] = 1;       // Detect all sprites over player fix

    vbuf = VL_LockSurface(screenBuffer);
    vbuf+=screenofs;
    vbufPitch = bufferPitch;

    CalcViewVariables();

//
// follow the walls from there to the right, drawing as we go
//
    VGAClearScreen ();
#if defined(USE_FEATUREFLAGS) && defined(USE_STARSKY)
    if(GetFeatureFlags() & FF_STARSKY)
        DrawStarSky(vbuf, vbufPitch);
#endif

    WallRefresh ();

#if defined(USE_FEATUREFLAGS) && defined(USE_PARALLAX)
    if(GetFeatureFlags() & FF_PARALLAXSKY)
        DrawParallax(vbuf, vbufPitch);
#endif
#if defined(USE_FEATUREFLAGS) && defined(USE_CLOUDSKY)
    if(GetFeatureFlags() & FF_CLOUDSKY)
        DrawClouds(vbuf, vbufPitch, min_wallheight);
#endif
#ifdef USE_FLOORCEILINGTEX
    //DrawFloorAndCeiling(vbuf, vbufPitch, min_wallheight);
#endif

//
// draw all the scaled images
//
    DrawScaleds();                  // draw scaled stuff

#if defined(USE_FEATUREFLAGS) && defined(USE_RAIN)
    if(GetFeatureFlags() & FF_RAIN)
        DrawRain(vbuf, vbufPitch);
#endif
#if defined(USE_FEATUREFLAGS) && defined(USE_SNOW)
    if(GetFeatureFlags() & FF_SNOW)
        DrawSnow(vbuf, vbufPitch);
#endif

    DrawPlayerWeapon ();    // draw player's hands

    if (Keyboard[sc_Tab] && viewsize == 21 && gamestate.weapon != -1) ShowActStatus();

    VL_UnlockSurface(screenBuffer);
    vbuf = NULL;

//
// show screen and time last cycle
//

    if (fizzlein)
    {
        FizzleFade(screenBuffer, 0, 0, screenWidth, screenHeight, 20, false);
        fizzlein = false;

        lasttimecount = GetTimeCount();          // don't make a big tic count
    }
    else
    {
#ifndef REMDEBUG
        if (fpscounter)
        {
            fontnumber = 0;
            SETFONTCOLOR(7,127);
            PrintX=4; PrintY=1;
            VWB_Bar(0,0,50,10,bordercol);
            US_PrintSigned(fps);
            US_Print(" fps");
        }
#endif
        SDL_BlitSurface(screenBuffer, NULL, screen, NULL);
        SDL_Flip(screen);
    }

#ifndef REMDEBUG
    if (fpscounter)
    {
        fps_frames++;
        fps_time+=tics;

        if(fps_time>35)
        {
            fps_time-=35;
            fps=fps_frames<<1;
            fps_frames=0;
        }
    }
#endif
}
